SET ANSI_NULLS, ANSI_PADDING, ANSI_WARNINGS, ARITHABORT, CONCAT_NULL_YIELDS_NULL, QUOTED_IDENTIFIER ON;
SET NUMERIC_ROUNDABORT OFF;
GO

IF (EXISTS (SELECT * 
                 FROM INFORMATION_SCHEMA.TABLES 
				 WHERE TABLE_SCHEMA = 'dbo' AND TABLE_NAME = 'aspnet_Membership'))
BEGIN
	IF NOT (EXISTS (SELECT * 
                 FROM INFORMATION_SCHEMA.TABLES 
				 WHERE TABLE_SCHEMA = 'dbo' AND TABLE_NAME = 'UserLogins'))
	BEGIN
		PRINT N'Creating table UserLogins...'
		CREATE TABLE [UserLogins] (
				[LoginProvider] NVARCHAR (128) NOT NULL,
				[ProviderKey]   NVARCHAR (128) NOT NULL,
				[UserId]        uniqueidentifier NOT NULL,
					CONSTRAINT [PK_dbo.UserLogins] PRIMARY KEY CLUSTERED ([LoginProvider] ASC, [ProviderKey] ASC, [UserId] ASC),
				CONSTRAINT [FK_dbo.UserLogins_dbo.aspnet_Users_UserId] FOREIGN KEY ([UserId]) REFERENCES [aspnet_Users] ([UserId]) ON DELETE CASCADE
			);
			CREATE NONCLUSTERED INDEX [IX_UserId]
				ON [UserLogins]([UserId] ASC);
	END
	ELSE
		PRINT N'The table UserLogins already present';

	IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = 'PersistedGrants')
	BEGIN
	    PRINT N'Creating PersistedGrants table ...';
		
		CREATE TABLE [PersistedGrants] (
			[Key] nvarchar(200) NOT NULL,
			[Type] nvarchar(50) NOT NULL,
			[SubjectId] nvarchar(200) NULL,
			[ClientId] nvarchar(200) NOT NULL,
			[CreationTime] datetime2 NOT NULL,
			[Expiration] datetime2 NULL,
			[Data] nvarchar(max) NOT NULL,
			[SessionId] nvarchar(100) NULL,
			[Description] nvarchar(200) NULL,
			[ConsumedTime] datetime2 NULL,
			CONSTRAINT [PK_PersistedGrants] PRIMARY KEY ([Key])
		);
	END
	ELSE
	BEGIN
		IF NOT EXISTS (
			SELECT * 
			FROM   sys.columns 
			WHERE  object_id = OBJECT_ID(N'[dbo].[PersistedGrants]') 
					AND name = 'SessionId'
		)
		BEGIN
		    PRINT N'Altering PersistedGrants table with adding SessionId colunm...';
			
			ALTER TABLE [PersistedGrants]
				ADD [SessionId] NVARCHAR (100) NULL
				PRINT N'Field [SessionId] added to [PersistedGrants] table';
		END

		IF NOT EXISTS (
			SELECT * 
			FROM   sys.columns 
			WHERE  object_id = OBJECT_ID(N'[dbo].[PersistedGrants]') 
					AND name = 'Description'
		)
		BEGIN
		    PRINT N'Altering PersistedGrants table with adding Description colunm...';
			
			ALTER TABLE [PersistedGrants]
				ADD [Description] NVARCHAR (200) NULL
				PRINT N'Field [Description] added to [PersistedGrants] table';
		END

		IF NOT EXISTS (
			SELECT * 
			FROM   sys.columns 
			WHERE  object_id = OBJECT_ID(N'[dbo].[PersistedGrants]') 
					AND name = 'ConsumedTime'
		)
		BEGIN
		    PRINT N'Altering PersistedGrants table with adding ConsumedTime colunm...';
		
			ALTER TABLE [PersistedGrants]
				ADD [ConsumedTime] datetime2 NULL
				PRINT N'Field [ConsumedTime] added to [PersistedGrants] table';
		END
	END
	
	IF NOT EXISTS(
		SELECT * 
		FROM sys.indexes 
		WHERE object_id = OBJECT_ID(N'[dbo].[PersistedGrants]') 
			AND name = 'IX_PersistedGrants_SubjectId_SessionId_Type'
	)
	BEGIN
		PRINT N'Creating Index [IX_PersistedGrants_SubjectId_SessionId_Type] for [PersistedGrants] table';
		
		CREATE INDEX [IX_PersistedGrants_SubjectId_SessionId_Type] ON [PersistedGrants] ([SubjectId] ASC, [SessionId] ASC, [Type] ASC);
		
		PRINT N'Index [IX_PersistedGrants_SubjectId_SessionId_Type] created for [PersistedGrants] table';
	END

	IF NOT EXISTS(
		SELECT * 
		FROM sys.indexes 
		WHERE object_id = OBJECT_ID(N'[dbo].[PersistedGrants]') 
			AND name = 'IX_PersistedGrants_SubjectId_ClientId_Type'
	)
	BEGIN
		PRINT N'Creating index [IX_PersistedGrants_SubjectId_ClientId_Type] for [PersistedGrants] table';
		
		CREATE INDEX [IX_PersistedGrants_SubjectId_ClientId_Type] ON [PersistedGrants] ([SubjectId] ASC, [ClientId] ASC, [Type] ASC);
		
		PRINT N'Index [IX_PersistedGrants_SubjectId_ClientId_Type] created for [PersistedGrants] table';
	END		

	IF NOT EXISTS(
		SELECT * 
		FROM sys.indexes 
		WHERE object_id = OBJECT_ID(N'[dbo].[PersistedGrants]') 
			AND name = 'IX_PersistedGrants_Expiration'
	)
	BEGIN
		PRINT N'Creating index [IX_PersistedGrants_Expiration] for [PersistedGrants] table';
		CREATE INDEX [IX_PersistedGrants_Expiration] ON [PersistedGrants] ([Expiration]);
		PRINT N'Index [IX_PersistedGrants_Expiration] created for [PersistedGrants] table';
	END		

	IF NOT EXISTS(
		SELECT * 
		FROM sys.indexes 
		WHERE object_id = OBJECT_ID(N'[dbo].[PersistedGrants]') 
			AND name = 'IX_PersistedGrants_ConsumedTime'
	)
	BEGIN
		PRINT N'Creating index [IX_PersistedGrants_ConsumedTime] for [PersistedGrants] table';
		
		CREATE INDEX [IX_PersistedGrants_ConsumedTime] ON [PersistedGrants] ([ConsumedTime]);
		
		PRINT N'Index [IX_PersistedGrants_ConsumedTime] created for [PersistedGrants] table';
	END	

	IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = 'DeviceCodes')
	BEGIN
		CREATE TABLE [DeviceCodes] (
			[UserCode] nvarchar(200) NOT NULL,
			[DeviceCode] nvarchar(200) NOT NULL,
			[SubjectId] nvarchar(200) NULL,
			[ClientId] nvarchar(200) NOT NULL,
			[CreationTime] datetime2 NOT NULL,
			[Expiration] datetime2 NOT NULL,
			[Data] nvarchar(max) NOT NULL,
			CONSTRAINT [PK_DeviceCodes] PRIMARY KEY ([UserCode])
		);
    
		CREATE UNIQUE INDEX [IX_DeviceCodes_DeviceCode] ON [DeviceCodes] ([DeviceCode]);
		CREATE INDEX [IX_DeviceCodes_Expiration] ON [DeviceCodes] ([Expiration]);
	END

	IF NOT EXISTS (SELECT * FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = 'ExternalUserData')
	BEGIN
		CREATE TABLE [ExternalUserData] (
			[Key] int IDENTITY(1,1),
			[ProviderName] nvarchar(200) NOT NULL,
			[UserId] nvarchar(200) NOT NULL,
			[IsActive] bit NOT NULL,
			[Data] nvarchar(max) NOT NULL,
			CONSTRAINT [PK_ExternalUserData] PRIMARY KEY ([Key])
		);

		CREATE UNIQUE INDEX [IX_ExternalUserData_ProviderName_UserId] ON [ExternalUserData] ([ProviderName], [UserId]);
	END
	
	IF NOT EXISTS (SELECT * FROM [dbo].[aspnet_Roles] WHERE [RoleName] = 'sitecore\Sitecore XConnect Data Admin')
	BEGIN
		PRINT N'Creating Sitecore XConnect Data Admin Role...'
		DECLARE @ApplicationId nvarchar(256) = 
			(SELECT [ApplicationId] 
			FROM [dbo].[aspnet_Applications]
			WHERE [ApplicationName] = 'sitecore');
		DECLARE @RoleId uniqueIdentifier = NEWID();
		
		INSERT INTO [dbo].[aspnet_Roles] (
			[ApplicationId],
			[RoleId],
			[RoleName],
			[LoweredRoleName])
		VALUES (
			@ApplicationId,
			@RoleId,
			'sitecore\Sitecore XConnect Data Admin',
			'sitecore\Sitecore xconnect data admin');
	END
END

-- Fix broken Shared fields (move them from VersionedFields table to SharedFields one)
INSERT INTO [SharedFields]
  SELECT [Id], [ItemId], [FieldId], [Value], [Created], [Updated] FROM [VersionedFields] Vf 
    WHERE [Vf].[FieldId] IN (SELECT [ItemId] FROM [SharedFields] WHERE [FieldId] = 'BE351A73-FCB0-4213-93FA-C302D8AB4F51' AND [Value] = '1')
    AND [Vf].[Language] = 'en' AND [Vf].[Version] = 1
    AND NOT EXISTS (SELECT 1 FROM [SharedFields] Sf WHERE [Sf].[ItemId] = [Vf].[ItemId] AND [Sf].[FieldId] = [Vf].[FieldId])

DELETE FROM [VersionedFields]
  WHERE [FieldId] IN (SELECT [ItemId] FROM [SharedFields] WHERE [FieldId] = 'BE351A73-FCB0-4213-93FA-C302D8AB4F51' AND [Value] = '1')
  AND [Language] = 'en' AND [Version] = 1

GO

PRINT N'Update complete.';